from django.db import models
from django.utils.six import with_metaclass
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError

from .subclassing import SubfieldBase
from .settings import kv_settings

serialize = kv_settings['SERIALIZE_FUNCTION']
deserialize = kv_settings['DESERIALIZE_FUNCTION']

class SerializableField(with_metaclass(SubfieldBase, models.TextField)):
    """
    A model field that restores it's type upon loading from the database.

    This field extends TextField so it's saved in the database as a text field.
    """

    def pre_init(self, value, obj):
        """
        deserializes the value saved after loading from the database,
        this function uses the DESERIALIZE_FUNCTION in the settings if one is
        provided.
        ValidationError is thrown if the value could not be deserialized.
        """
        if obj._state.adding:
            # Make sure the primary key actually exists on the object before
            # checking if it's empty. This is a special case for South datamigrations
            # this is also inspired by
            # https://github.com/bradjasper/django-jsonfield/
            if not hasattr(obj, "pk") or obj.pk is not None:
                try:
                    return deserialize(value)
                except ValueError:
                    raise ValidationError(_("could not deserialize the given value."))

        return value

    def to_python(self, value):
        return value

    def get_prep_value(self, value):
        return self.value_to_string(value)

    def value_to_string(self, obj):
        """
        serializes object before saving, this function uses the
        SERIALIZE_FUNCTION in settings if one is provided.
        ValidationError is thrown if the value could not be serialized.
        """
        try:
            return serialize(obj)
        except ValueError:
            raise ValidationError(_("could not serialize the given value."))

    def __unicode__(self):
        return value_to_string(self.value)



"""
the default serializer for the plugin.

it uses the json to do the actual work.
"""

import json

def json_serialize(value):
    return json.dumps(value)

def json_deserialize(s):
    return json.loads(s)



"""
Example
-------

class SystemSetting(KVModel):
    pass

setting = SystemSetting.create(key='foo', value=100)
loaded_setting = SystemSetting.get_by_key('foo')

"""

from django.db import models

from .fields import SerializableField

class KVModel(models.Model):
    """
    An Abstract model that has key and value fields

    key   -- Unique CharField of max_length 255
    value -- SerializableField by default could be used to store bool, int,
             float, str, list, dict and date
    """

    key = models.CharField(max_length=255, unique=True)
    value = SerializableField(blank=True, null=True)

    def __unicode__(self):
        return 'KVModel instance: ' + self.key + ' = ' + unicode(self.value)

    @classmethod
    def get_by_key(cls, key):
        """
        A static method that returns a KVModel instance.

        key -- unique key that is used for the search.
        this method will throw a DoesNotExist exception if an object with the
        key provided is not found.
        """
        return cls.objects.get(key=key)

    class Meta:
        abstract = True



from django.test import TestCase
from django.db import models

from .serializers import json_serialize, json_deserialize
from .models import KVModel
from .fields import SerializableField

class KVModelConcrete(KVModel):
    """
    a class that extends KVModel and is used
    for testing purposses only
    """

class ModelWithSerializableField(models.Model):
    """
    this class is used to test the serialzible field
    """

    value = SerializableField()

class TestSerializer(TestCase):

    def test_serialize_desialize_int(self):
        value = 12
        serialized = '12'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_desialize_boolean(self):
        value = True
        serialized = 'true'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_desialize_string(self):
        value = 'True'
        serialized = '"True"'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_desialize_float(self):
        value = 12.32
        serialized = '12.32'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_desialize_empty_string(self):
        value = ''
        serialized = '""'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_desialize_empty_zero(self):
        value = 0
        serialized = '0'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_desialize_array(self):
        value = [1, 3, 'man', True]
        serialized = '[1, 3, "man", true]'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_desialize_empty_array(self):
        value = []
        serialized = '[]'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_None(self):
        value = None
        serialized = 'null'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_empty_dict(self):
        value = {}
        serialized = '{}'
        self.assertEquals(json_serialize(value), serialized)
        self.assertEquals(json_deserialize(serialized), value)

    def test_serialize_dict(self):
        value = {'a': 123, 'b': '234'}
        self.assertEquals(json_deserialize(json_serialize(value)), value)

class TestSerializableField(TestCase):

    def test_int(self):
        value = 100000
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_string(self):
        value = '"";string"'
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_array(self):
        value = ['man', 'dragon']
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_empty_array(self):
        value = []
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_empty_string(self):
        value = ''
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_none(self):
        value = None
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_float(self):
        value = 23.43
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_zero_float(self):
        value = 0.0
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_dict(self):
        value = {'foo': 'bar', 'man': 0}
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

    def test_empyty_dict(self):
        value = {}
        obj = ModelWithSerializableField(value=value)
        obj.save()
        self.assertEquals(value, obj.value)

        obj = ModelWithSerializableField.objects.first()
        self.assertEquals(value, obj.value)

class TestKVModel(TestCase):

    def test_creating_with_a_string_value(self):
        key = 'foo'
        value = '"bar"'
        obj = KVModelConcrete(key=key, value=value)
        obj.save()
        obj = KVModelConcrete.get_by_key('foo')
        self.assertEquals(obj.value, value)

    def test_creating_with_an_int_value(self):
        key = 'foo'
        value = 1999
        obj = KVModelConcrete(key=key, value=value)
        obj.save()
        obj = KVModelConcrete.get_by_key(key=key)
        self.assertEquals(obj.value, value)

    def test_creating_with_an_int_value(self):
        key = 'foo'
        value = 1999
        obj = KVModelConcrete(key=key, value=value)
        obj.save()
        obj = KVModelConcrete.get_by_key(key=key)
        self.assertEquals(obj.value, value)

    def test_creating_with_a_float_value(self):
        key = 'foo'
        value = 1999.00
        obj = KVModelConcrete(key=key, value=value)
        obj.save()
        obj = KVModelConcrete.get_by_key(key=key)
        self.assertEquals(obj.value, value)

    def test_creating_with_a_None_value(self):
        key = 'foo'
        value = None
        obj = KVModelConcrete(key=key, value=value)
        obj.save()
        obj = KVModelConcrete.get_by_key(key=key)
        self.assertEquals(obj.value, value)



import importlib

from django.conf import settings

def import_function(function_path):
    module_name, function_name = function_path.rsplit('.',1)
    module = importlib.import_module(module_name)
    return getattr(module, function_name)

DEFAULTS = {
    'SERIALIZE_FUNCTION': 'kvmodel.serializers.json_serialize',
    'DESERIALIZE_FUNCTION': 'kvmodel.serializers.json_deserialize',
}

USER_SETTINGS = getattr(settings, 'KVMODEL', {})

DEFAULTS.update(USER_SETTINGS)

kv_settings = {k: import_function(DEFAULTS[k]) for k in DEFAULTS}



from setuptools import setup

description = 'A package that helps creating django key-value models easily.'

try:
    with open('README.rst') as f:
        long_description = f.read()
except IOError:
    long_description = description

setup(
    name='django-kvmodel',
    version='0.1.0',
    description=description,
    author='Abdallah Dorra',
    author_email='amdorra@gmail.com',
    url='https://github.com/amdorra/django-kvmodel',
    license='LICENSE.txt',
    long_description=long_description,
    packages=['kvmodel'],
    install_requires=['django >= 1.2.7'],
)
